/*
 * gdb-stubd
 *
 * A daemon offering gdb services sporting a variety of back ends.
 *
 * (c) 1999 Austin Donnelly <Austin.Donnelly@cl.cam.ac.uk>
 *
 * You may distribute copies of this program under the GPL.  There is
 * NO WARRANTY associated with this software.
 *
 */

#ifndef backend_h
#define backend_h

/*
 * Operations provided by the backend
 */


/* This is called before any other backend function is called, and it
 * returns an opaque value which is passed in as the first argument of
 * all future backend calls. */
typedef void * (*be_init_fn) (void);

/* This (optional) function is called on the backed to let it know
 * about any remaining arguments on the command line. */
typedef void (*be_process_args_fn) (void *, int optind, int argc, char **argv);

/* Optional function that's called when client disconnects.  See also
 * the "kill" function.  */
typedef void (*be_finished_fn) (void *);

/* Return which signal number the process died from */
typedef int (*be_sigval_fn) (void *);

/* Return a freshly allocated block of memory describing the register
 * contents, along with the number of bytes in the block. */
typedef uint8_t * (*be_regdump_fn) (void *, int *regdumpbytes);

/* Read "numbytes" from "addr" onwards, filling "buf" with the data.
 * Return TRUE for success, FALSE otherwise. */
typedef bool_t (*be_memread_fn) (void *, uint8_t *buf, int addr, int numbytes);

/* Write "numbytes" from "buf" to "addr".
 * Return TRUE for success, FALSE otherwise. */
typedef bool_t (*be_memwrite_fn)(void *, uint8_t *buf, int addr, int numbytes);

/* Optional.  Called when GDB asks to kill the process being debugged. */
typedef void (*be_kill_fn) (void *);

/* Optional.
 * Execute a single instruction, starting at "addr" if "addr_valid" is
 * TRUE, otherwise starting at the current PC.  Returns the signal we
 * stopped with. */
typedef int (*be_step_fn) (void *, int addr, bool_t addr_valid);

/* Optional.
 * Execute, returning (when the process has stopped) with a signal number 
 * Start at "addr" if "addr_valid" is true, otherwise at current PC. */
/* XXX ought to thread this off so we can take requestsw to interrupt? */
typedef int (*be_cont_fn) (void *, int addr, bool_t addr_valid);

typedef struct {
    be_init_fn		init;
    be_process_args_fn	process_args;
    be_finished_fn	finished;
    be_sigval_fn	sigval;
    be_regdump_fn	regdump;
    be_memread_fn	memread;
    be_memwrite_fn	memwrite;
    be_kill_fn		kill;
    be_step_fn		step;
    be_cont_fn          cont;
} backend_ops; 


typedef struct {
    const char	*name;
    backend_ops	*ops;
} backend_info;



#endif /* backend_h */
