/*
 * gdb-stubd
 *
 * A daemon offering gdb services sporting a variety of back ends.
 *
 * This backend (c) 1999 Tim Deegan <Tim.Deegan@cl.cam.ac.uk>
 *
 * You may distribute copies of this program under the GPL.  There is
 * NO WARRANTY associated with this software.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <linux/types.h>

#include "util.h"
#include "backend.h"
#include "acenic_ioctls.h"

static void *   alteon_init (void);
static void     alteon_process_args (void*, int optind, int argc, char **argv);
static void     alteon_finished (void *);
static int      alteon_sigval (void *);
static uint8_t *alteon_regdump (void *, int *regdumpbytes);
static bool_t   alteon_memread (void *, uint8_t *buf, int addr, int numbytes);
static bool_t   alteon_memwrite (void *, uint8_t *buf, int addr, int numbytes);
static void     alteon_kill (void *);
static int      alteon_step (void *, int addr, bool_t addr_valid);
static int      alteon_cont (void *, int addr, bool_t addr_valid);

backend_ops be_alteon_ops = {
    alteon_init,
    alteon_process_args,
    alteon_finished,
    alteon_sigval,
    alteon_regdump,
    alteon_memread,
    alteon_memwrite,
    alteon_kill,
    alteon_step,
    alteon_cont
};

/**************************************************************/

typedef struct {
    int socket;
    int cpu;
    struct ifreq ifr;
    struct acenic_ioc_req req;
} alteon_st;


/* 90 32-bit registers.  The Tigon only actually has 31... */
typedef struct {
    uint32_t gpr[32];
    uint32_t pad1[5];
    uint32_t pc;
    uint32_t pad2[34];
    uint32_t fp; /* == sp == gpr[30] on the Tigon */
    uint32_t pad3[17];
} alteon_regs;

/**************************************************************/
/* 
 *  Helper functions 
 */

/* MIPS is big-endian but PCI but gives us little-endian values, so
 * the ioctl routines need to switch them to big_endian so GDB gets
 * the contents right.  We can use ntohl, htonl to access control
 * register contents, &c. here in the stub.
 *
 * NB: 'offset' and card addresses are _host-endian_ and do not need htonl().
 */

static __inline__ uint32_t 
other_endian (uint32_t val)
{
    return (((val & 0xff) << 24)
	    | ((val & 0xff00) << 8)
	    | ((val & 0xff0000) >> 8)
	    | ((val & 0xff000000) >> 24));
}

/* Read and write shared memory locations */

static bool_t
read_smem (alteon_st *nic, uint32_t offset, uint32_t *dpr)
{
    nic->req.cardoffset = offset;
    if (ioctl(nic->socket, ACENIC_IOCTL_READ_SMEM, &nic->ifr))
    {
	if (verbose > 2)
	{
	    printf ("read ioctl failed: errno  = %i (%s)\n"
		    "                   offset = %#lx\n"
		    "                   data   = %#x\n",
		    errno, strerror(errno),
		    nic->req.cardoffset, 
		    ntohl(other_endian(nic->req.data)));
	}
	*dpr = other_endian(nic->req.data);
	return FALSE;
    }
    *dpr = other_endian(nic->req.data);
    return TRUE;
}

static bool_t 
write_smem (alteon_st *nic, uint32_t offset, uint32_t data)
{
    nic->req.cardoffset = offset;
    nic->req.data = other_endian(data);
    if (verbose > 2)
    {
	printf ("write ioctl: offset = %#lx\n"
		"             data   = %#x\n",
		nic->req.cardoffset, 
		ntohl(other_endian(nic->req.data)));
    }
    if (ioctl(nic->socket, ACENIC_IOCTL_WRITE_SMEM, &nic->ifr))
    {
        if (verbose > 2) 
	{
	    printf ("write ioctl failed: errno is %i (%s)\n", 
		    errno, strerror(errno));
	}
	return FALSE;
    }

    if (verbose > 2) 
    {
	printf("             result = %#x\n", 
	       ntohl(other_endian(nic->req.data)));
    }
    return TRUE;
}

/* Read and write card addresses (translate to shared memory locations) */

static bool_t
nic_read_32 (alteon_st *nic, uint32_t addr, uint32_t *dpr)
{
    uint32_t offset = addr;
    uint32_t save, rv;
    /* How we do this depends on what the address is */
    if (offset < 0x800000) {
	/* NIC RAM -- use 'window base address' and 'window data' regs */
	if (!write_smem (nic, 0x68, htonl(offset)))
	    return (FALSE);
	return (read_smem (nic, 0x6C, dpr));
    } else if (offset >= 0xC00000 && offset < 0xC04000) {	
	/* CPU scratchpad -- use internal SRAM indirect registers */
	offset -= 0xC00000;
	/* First halt the CPU */
	if ((!read_smem (nic, 0x140 + (nic->cpu * 0x100), &save))	   
	    || (!write_smem (nic, 0x140 + (nic->cpu * 0x100), 
			     save | htonl(0x10000)))
	    /* Read the value */
	    || (!write_smem (nic, 0x154 + (nic->cpu * 0x100), htonl(offset)))
	    || (!read_smem (nic, 0x158 + (nic->cpu * 0x100), &rv))
	    /* Put the CPU state back */
	    || (!write_smem (nic, 0x140 + (nic->cpu * 0x100), save)))
	    return FALSE;
	*dpr = rv;
	return TRUE;
    } else if (offset >= 0xC0000000 && offset < 0xC0000400) {
	/* HW registers -- direct mapped*/
	offset -= 0xC0000000;
	return (read_smem (nic, offset, dpr));
    } else {
	/* Invalid */
	return FALSE;
    }   
}

static bool_t
nic_write_32 (alteon_st *nic, uint32_t addr, uint32_t data)
{
    uint32_t offset = addr;
    uint32_t save;
    /* How we do this depends on what the address is */
    if (offset < 0x800000) {
	/* NIC RAM -- use 'window base address' and 'window data' regs */
	if (!write_smem (nic, 0x68, htonl(offset)))
	    return (FALSE);
	return (write_smem (nic, 0x6C, data));
    } else if (offset >= 0xC00000 && offset < 0xC04000) {	
	/* CPU scratchpad -- use internal SRAM indirect registers */
	offset -= 0xC00000;
	/* First halt the CPU */
	if ((!read_smem (nic, 0x140 + (nic->cpu * 0x100), &save))	   
	    || (!write_smem (nic, 0x140 + (nic->cpu * 0x100), 
			     save | htonl(0x10000)))
	    /* Write the value */
	    || (!write_smem (nic, 0x154 + (nic->cpu * 0x100), htonl(offset)))
	    || (!write_smem (nic, 0x158 + (nic->cpu * 0x100), data))
	    /* Put the CPU state back */
	    || (!write_smem (nic, 0x140 + (nic->cpu * 0x100), save)))
	    return FALSE;
	return TRUE;
    } else if (offset >= 0xC0000000 && offset < 0xC0000400) {
	/* HW registers -- direct mapped*/
	offset -= 0xC0000000;
	return (write_smem (nic, offset, data));
    } else {
	/* Invalid */
	return FALSE;
    }   
}

static int
state2sig (uint32_t state)
{
    /* Pick an appropriate signal number depending on the state register 
     * The numbers are defined in target.h in GDB 4.16 */
    
    state = htonl(state);  /* Let's talk host-endian for now */
    
    if (state & 0x10000)   /* External halt bit */
	return 2;          /* SIGINT */
    if (state & 0x20000)   /* HW breakpoint */
	return 5;          /* SIGTRAP */
    if (state & 0x40000)   /* Halt instrn executed */
	return 6;          /* SIGABRT */
    if (state & 0x80000)   /* Invalid instruction */
	return 4;          /* SIGILL */
    if (state & 0x300000)  /* Page 0 data reference or page 0 instrn fetch */
	return 11;         /* SIGSEGV */
    if (state & 0x1c00000) /* Invalid data/instrn fetch or bad alignment */
	return 10;         /* SIGBUS */

    /* Default - means we called state2sig on a running CPU */
    return 0;
}

/**************************************************************/

static void *
alteon_init (void)
{
    alteon_st *st;

    if (geteuid() != 0) 
	die ("Alteon backend needs to be run as root");

    st = malloc (sizeof (*st));
    if (!st)
	die ("out of memory allocating alteon state record");

    memset (st, 0, sizeof (*st));
    
    /* Open a socket for our ioctls */
    st->socket = socket(AF_INET, SOCK_DGRAM, IPPROTO_IP);
    st->ifr.ifr_data = (char *)&st->req;
    st->cpu = -1;
    
    return st;
}


static void
alteon_process_args (void *data, int optind, int argc, char **argv)
{
    alteon_st *st = data;
    int cpu = 0;

    /* Args: 0 = interface name, 1 = cpu (optional) */
    
    /* Sanity check */
    if ((argc - optind < 1)
	|| (strlen(argv[optind]) >= IFNAMSIZ))
	return;
    strcpy(st->ifr.ifr_name, argv[optind]);

    if (argc - optind > 1)
	cpu = atoi(argv[optind + 1]);
    if (cpu & ~1)
	return;
    
    st->cpu = cpu;
    
    if (verbose) 
	printf ("%s: Alteon backend on interface %s, cpu %i\n",
		progname, st->ifr.ifr_name, st->cpu);
    
    return;
}

static void
alteon_finished (void *data)
{
//    alteon_st *st = data;

    /* do we need to do anything special when the client goes away? */

    return;
}

/* What signal did we die from? */
static int
alteon_sigval (void *be)
{
    return 0;
}


static uint8_t *
alteon_regdump (void *data, int *regdumpbytes)
{
    /* No point in doing preoper error returns - if we can't read the 
     * register set it's all gone wrong */
    alteon_st *st = data;
    alteon_regs *regs;
    uint32_t offset;
    int i;

    regs = malloc (sizeof (*regs));
    if (!regs)
	die ("malloc failed allocing %d bytes for alteon_regs",
	     sizeof (*regs));
    
    memset (regs, 0, sizeof (*regs));
    
    if (!(st->cpu & ~1))  /* Valid state */
    {    	
	/* PC first */
	offset = 0x144 + (0x100 * st->cpu);
	read_smem (st, offset, &regs->pc);
	
	/* GPRs */
	offset = 0x180 + (0x100 * st->cpu);
	for (i=0; i<32; i++)
	    read_smem (st, offset + (i * 4), &regs->gpr[i]);
	
	/* FP special case */
	regs->fp = regs->gpr[30];	
    }    
    
    (*regdumpbytes) = sizeof (*regs);
    return (void*)regs;
}


static bool_t
alteon_memread (void *data, uint8_t *buf, int addr, int numbytes)
{
    alteon_st *st = data;
    uint32_t offset, tmp;
    uint32_t *dest, *limit;
    int i;

    if ((st->cpu & ~1))  /* not connected */
    {
	memset(buf, 0, numbytes);
	return FALSE;
    } else {

	/* Fill in buf with numbytes of data from addr */ 
	
	offset = ((addr + 3) & ~0x3);
	dest = (uint32_t *)((buf + (offset - addr)));
	limit = (dest + ((numbytes - (offset - addr)) >> 2));
	
	/* tidy up initial few bytes */
	if (addr & 0x3)
	{
	    if (!nic_read_32 (st, (addr & ~0x3), &tmp))
		return FALSE;
	    for (i = (addr & 0x3); i < 3; i++)
	    {
		((char *)dest)[i - 4] = ((char *)(&tmp))[i];	    
	    }
	}
	
	/* main loop */
	while (dest < limit)
	{
	    if (!nic_read_32 (st, offset, dest))
		return FALSE;
	    offset += 4;
	    dest += 1;
	}
	
	/* tidy up top end */
	if (((addr + numbytes) & ~0x3))
	{
	    if (!nic_read_32 (st, offset, &tmp))
		return FALSE;
	    for (i=0; i < ((addr + numbytes) & 0x3); i++)
	    {
		((char *)dest)[i] = ((char *)(&tmp))[i];	    
	    }
	}
	return TRUE;
    }
}

static bool_t
alteon_memwrite (void *data, uint8_t *buf, int addr, int numbytes)
{
    alteon_st *st = data;
    uint32_t offset, tmp;
    uint32_t *dest, *limit;
    int i;

    if ((st->cpu & ~1))  /* not connected */
    {
	return FALSE;
    } else {

	/* Fill addr with numbytes of data from buf */
	
	offset = (addr + 3) & ~0x3;
	dest = (uint32_t *)((buf + (offset - addr)));
	limit = (dest + ((numbytes - (offset - addr)) >> 2));
	
	/* tidy up initial few bytes */
	if (addr & 0x3)
	{
	    if (!nic_read_32 (st, (addr & ~0x3), &tmp))
		return FALSE;
	    for (i = (addr & 0x3); i < 3; i++)
	    {
		((char *)(&tmp))[i] = ((char *)dest)[i - 4];
	    }
	    if (!nic_write_32 (st, (addr & ~0x3), tmp))
		return FALSE;
	}
	
	/* main loop */
	while (dest < limit)
	{
	    if (!nic_write_32 (st, offset, *dest))
		return FALSE;
	    offset += 4;
	    dest += 1;
	}
	
	/* tidy up top end */
	if (((addr + numbytes) & ~0x3))
	{
	    if (!nic_read_32 (st, offset, &tmp))
		return FALSE;
	    for (i=0; i < ((addr + numbytes) & 0x3); i++)
	    {
		((char *)(&tmp))[i] = ((char *)dest)[i];
	    }
	    if (!nic_write_32 (st, offset, tmp))
		return FALSE;
	}
	
	return TRUE;
    }
}

static int
alteon_step (void *data, int addr, bool_t addr_valid)
{
    /* Single-step the CPU, starting at 'addr' if 'addr_valid' is TRUE */
    alteon_st *st = data;
    uint32_t offset, state;

    if ((st->cpu & ~1))  /* not connected */
	return 1;

    /* Paranoia: make sure the CPU is halted... */
    offset = 0x140 + (0x100 * st->cpu);  /* CPU state register */
    read_smem(st, offset, &state);
    state |= htonl(0x10000);             /* "Halt CPU" bit */
    write_smem (st, offset, state);
    
    /* Do we need to set the PC? */
    if (addr_valid) 
    {
	offset = 0x144 + (0x100 * st->cpu);  /* PC */	
	write_smem (st, offset, (uint32_t) htonl(addr));
    }

    offset = 0x140 + (0x100 * st->cpu);  /* CPU state register */
    read_smem(st, offset, &state);
    state |= htonl(0x8002);  /* Write-only single-step bit, flush I-cache */
    write_smem (st, offset, state);

    /* Re-get the state to check the signal number */
    read_smem(st, offset, &state);
    return state2sig(state);
}

static int
alteon_cont (void *data, int addr, bool_t addr_valid)
{
    /* Continue, and watch the CPU till it stops */
    
    alteon_st *st = data;
    uint32_t offset, state;

    if ((st->cpu & ~1))  /* not connected */
	return 1;
    
    /* Read the state */
    offset = 0x140 + (0x100 * st->cpu);  /* CPU state register */
    read_smem(st, offset, &state);

    /* Do we need to set the PC? */
    if (addr_valid) 
    {
	/* Paranoia: make sure the CPU is halted... */
	state |= htonl(0x10000);  /* "Halt CPU" bit */
	write_smem (st, offset, state);
    
	write_smem (st, (offset + 4), (uint32_t) htonl(addr));  /* Write PC */
    }
    
    /* Start the CPU */
    state &= htonl(0x1ff);  
    /* Top 2 bytes are halt conditions, also don't want watchdog on */
    state |= htonl(0x8000); /* Flush I-cache */
    write_smem (st, offset, state);
    
    /* Poll */
    do {
	sleep(1);
	read_smem(st, offset, &state);
	if (verbose > 1) printf(".");
    } while (!(state & htonl(0xffff0000))); 
    
    return state2sig(state);
}

static void
alteon_kill (void *data)
{
    alteon_st *st = data;

    /* disconnect etc... */
    st->cpu = -1;
    close(st->socket);

    return;
}
