/*
 * gdb-stubd
 *
 * A daemon offering gdb services sporting a variety of back ends.
 *
 * (c) 1999 Austin Donnelly <Austin.Donnelly@cl.cam.ac.uk>
 *
 * You may distribute copies of this program under the GPL.  There is
 * NO WARRANTY associated with this software.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

#include <netdb.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#include "util.h"
#include "backend.h"
#include "stub.h"

#define DEFAULT_PORT htons(5000)
#define ADDR_LEN sizeof(struct sockaddr_in)

/* local prototypes */
static backend_ops *backend_ops_by_name (const char *name);
static void usage (void);
static int start_listening (const char *hostname, uint16_t port);


/* globals */
const char *progname;
int verbose = 0;

/* table of all backends */
extern backend_ops be_alteon_ops;
backend_info backend[] = {
    /* the first backend is the default one */
    { "alteon", &be_alteon_ops },
};
#define NUM_BACKENDS ((sizeof (backend) / sizeof (backend_info)))


/**************************************************************/
/* Main */

int
main (int argc, char **argv)
{
    int ret;
    int i;
    backend_ops *cur_ops = backend[0].ops;
    void *be;
    char *hostname = NULL;
    int clientfd;
    uint16_t port = DEFAULT_PORT;

    /* basename our executable's name */
    progname = strrchr (argv[0], '/');
    if (progname)
	progname++;
    else
	progname = argv[0];

    /* parse our options */
    while ((ret = getopt (argc, argv, "b:v")) != EOF)
    {
	switch (ret) {
	case 'v':
	    verbose++;
	    break;

	case 'b':	/* specify a particular backend */
	    cur_ops = backend_ops_by_name (optarg);
	    if (!cur_ops)
	    {
		fprintf(stderr,
			"%s: unknown backend '%s'. Valid backends are:\n",
			progname, optarg);
		for (i=0; i < NUM_BACKENDS; i++)
		    fprintf(stderr, "\t%s\n", backend[i].name);
		exit(1);
	    }
	    break;

	case '?':	/* bogus option given */
	    usage ();
	    break;

	case ':':
	    die ("'%c' option needs a parameter", optopt);
	    break;

	default:
	    die ("getopt() returned weird value %d", ret);
	    break;
	}
    }

    /* do we have a "[hostname:]port" argument? */
    if (optind < argc)
    {
	char *portnum;
	portnum = strchr (argv[optind], ':');
	if (portnum)
	{
	    hostname = argv[optind];
	    *portnum = '\000';
	    portnum++;
	}
	else
	{
	    hostname = NULL;
	    portnum = argv[optind];
	}

	ret = atoi (portnum);
	if (!ret)
	    die ("port argument must be an integer");
	if (ret < 1 || ret > 0xffff)
	    die ("port number must be between 0 - 65536");
	port = htons ((uint16_t)ret);

	optind++;
    }

    /* initialise the backend */
    be = NULL;
    if (cur_ops->init)
	be = cur_ops->init ();

    /* let the backend grab any remaining args if it wants to */
    if (optind < argc && cur_ops->process_args)
	cur_ops->process_args (be, optind, argc, argv);

    /* wait for a client to connect */
    clientfd = start_listening (hostname, port);

    stub_loop (clientfd, cur_ops, be);

    close (clientfd);

    if (cur_ops->finished)
	cur_ops->finished (be);

    return 0;
}

static void
usage (void)
{
    fprintf (stderr, "usage: %s [-b backend] [-v] [[hostname:]port] [...]\n",
	     progname);
    fprintf(stderr,"\t-b backend specifies a different backend (default=%s)\n",
	    backend[0].name);
    fprintf (stderr, "\t-v turns up the verbosity\n");
    fprintf (stderr, "\tport specifies which port to listen on\n");
    fprintf (stderr,
	     "\thostname can be used to select a particular interface\n");
    fprintf (stderr,
	     "\tany further arguments may be interpreted by the backend\n");
    exit (2);
}


static backend_ops *
backend_ops_by_name (const char *name)
{
    int i;

    for (i=0; i < NUM_BACKENDS; i++)
	if (!strcmp (name, backend[i].name))
	    return backend[i].ops;

    return NULL;
}



/*************************************************************/
/* Socket operations */


static int
start_listening (const char *hostname, uint16_t port)
{
    int listenfd;
    int clientfd;
    int raddrlen;
    struct sockaddr_in laddr, raddr;

    laddr.sin_family = AF_INET;
    laddr.sin_port   = port;

    /* if we have a hostname, then try and bind to just that interface */
    if (hostname)
    {
	struct hostent *hent;
	hent = gethostbyname (hostname);
	if (!hent)
	{
	    fprintf (stderr, "%s: ", progname);
	    herror (hostname);
	    exit (3);
	}
	if (hent->h_length != 4)
	    die ("address length is %d, was expecting 4 bytes!",
		 hent->h_length);
	memcpy (&laddr.sin_addr.s_addr, hent->h_addr, 4);
    }
    else
    {
	laddr.sin_addr.s_addr = INADDR_ANY;
    }


    listenfd = socket (laddr.sin_family, SOCK_STREAM, IPPROTO_TCP);
    if (listenfd < 0)
	die_errno ("listen socket");

    if (bind (listenfd, (struct sockaddr *)&laddr, ADDR_LEN) < 0)
	die_errno ("listen bind");

    if (listen (listenfd, 1) < 0)
	die_errno ("listen");

    if (verbose)
	printf ("%s: listening on %s:%d...\n",
		progname, inet_ntoa (laddr.sin_addr), (int)ntohs (port));

    /* wait for a call to come in */
    raddrlen = ADDR_LEN;
    clientfd = accept (listenfd, (struct sockaddr*)&raddr, &raddrlen);
    if (clientfd < 0)
	die_errno ("accept");
    if (raddrlen != ADDR_LEN)
	die ("remote address is %d bytes long, was expecting 4!", raddrlen);

    if (verbose)
	printf ("Connect from %s:%d\n",
		inet_ntoa (raddr.sin_addr), (int)ntohs (raddr.sin_port));

    /* we only handle a single client */
    close (listenfd);

    return clientfd;
}
