/*
 * gdb-stubd
 *
 * A daemon offering gdb services sporting a variety of back ends.
 *
 * (c) 1999 Austin Donnelly <Austin.Donnelly@cl.cam.ac.uk>
 * Portions of this code from i386-stub.c from the GDB distribution.
 *
 * You may distribute copies of this program under the GPL.  There is
 * NO WARRANTY associated with this software.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

#include "util.h"
#include "backend.h"
#include "stub.h"


/* From the GDB distribution: */
/* Remote communication protocol.

   A debug packet whose contents are <data>
   is encapsulated for transmission in the form:

	$ <data> # CSUM1 CSUM2

	<data> must be ASCII alphanumeric and cannot include characters
	'$' or '#'.  If <data> starts with two characters followed by
	':', then the existing stubs interpret this as a sequence number.

	CSUM1 and CSUM2 are ascii hex representation of an 8-bit 
	checksum of <data>, the most significant nibble is sent first.
	the hex digits 0-9,a-f are used.

   Receiver responds with:

	+	- if CSUM is correct and ready for next packet
	-	- if CSUM is incorrect

   <data> is as follows:
   Most values are encoded in ascii hex digits.  Signal numbers are according
   to the numbering in target.h.

	Request		Packet

	set thread	Hct...		Set thread for subsequent operations.
					c = 'c' for thread used in step and 
					continue; t... can be -1 for all
					threads.
					c = 'g' for thread used in other
					operations.  If zero, pick a thread,
					any thread.
	reply		OK		for success
			ENN		for an error.

	read registers  g
	reply		XX....X		Each byte of register data
					is described by two hex digits.
					Registers are in the internal order
					for GDB, and the bytes in a register
					are in the same order the machine uses.
			or ENN		for an error.

	write regs	GXX..XX		Each byte of register data
					is described by two hex digits.
	reply		OK		for success
			ENN		for an error

        write reg	Pn...=r...	Write register n... with value r...,
					which contains two hex digits for each
					byte in the register (target byte
					order).
	reply		OK		for success
			ENN		for an error
	(not supported by all stubs).

	read mem	mAA..AA,LLLL	AA..AA is address, LLLL is length.
	reply		XX..XX		XX..XX is mem contents
					Can be fewer bytes than requested
					if able to read only part of the data.
			or ENN		NN is errno

	write mem	MAA..AA,LLLL:XX..XX
					AA..AA is address,
					LLLL is number of bytes,
					XX..XX is data
	reply		OK		for success
			ENN		for an error (this includes the case
					where only part of the data was
					written).

	continue	cAA..AA		AA..AA is address to resume
					If AA..AA is omitted,
					resume at same address.

	step		sAA..AA		AA..AA is address to resume
					If AA..AA is omitted,
					resume at same address.

	continue with	Csig;AA..AA	Continue with signal sig (hex signal
	signal				number).  If ;AA..AA is omitted, resume
					at same address.

	step with	Ssig;AA..AA	Like 'C' but step not continue.
	signal

	last signal     ?               Reply the current reason for stopping.
                                        This is the same reply as is generated
					for step or cont : SAA where AA is the
					signal number.

	detach          D               Reply OK.

	There is no immediate reply to step or cont.
	The reply comes when the machine stops.
	It is		SAA		AA is the signal number.

	or...		TAAn...:r...;n...:r...;n...:r...;
					AA = signal number
					n... = register number (hex)
					  r... = register contents
					n... = `thread'
					  r... = thread process ID.  This is
						 a hex integer.
					n... = other string not starting 
					    with valid hex digit.
					  gdb should ignore this n,r pair
					  and go on to the next.  This way
					  we can extend the protocol.
	or...		WAA		The process exited, and AA is
					the exit status.  This is only
					applicable for certains sorts of
					targets.
	or...		XAA		The process terminated with signal
					AA.
        or...           OXX..XX	XX..XX  is hex encoding of ASCII data. This
					can happen at any time while the program is
					running and the debugger should
					continue to wait for 'W', 'T', etc.

	thread alive	TXX		Find out if the thread XX is alive.
	reply		OK		thread is still alive
			ENN		thread is dead
	
	remote restart	RXX		Restart the remote server

	extended ops 	!		Use the extended remote protocol.
					Sticky -- only needs to be set once.

	kill request	k

	toggle debug	d		toggle debug flag (see 386 & 68k stubs)
	reset		r		reset -- see sparc stub.
	reserved	<other>		On other requests, the stub should
					ignore the request and send an empty
					response ($#<checksum>).  This way
					we can extend the protocol and GDB
					can tell whether the stub it is
					talking to uses the old or the new.
	search		tAA:PP,MM	Search backwards starting at address
					AA for a match with pattern PP and
					mask MM.  PP and MM are 4 bytes.
					Not supported by all stubs.

	general query	qXXXX		Request info about XXXX.
	general set	QXXXX=yyyy	Set value of XXXX to yyyy.
	query sect offs	qOffsets	Get section offsets.  Reply is
					Text=xxx;Data=yyy;Bss=zzz

	Responses can be run-length encoded to save space.  A '*' means that
	the next character is an ASCII encoding giving a repeat count which
	stands for that many repititions of the character preceding the '*'.
	The encoding is n+29, yielding a printable character where n >=3 
	(which is where rle starts to win).  Don't use an n > 126.

	So 
	"0* " means the same as "0000".  */


/* local prototypes */
static int      hex (char ch);
static uint8_t *getpacket (uint8_t *buffer, int buflen, int requestfd);
static void     putpacket (uint8_t *buffer, int requestfd);
static uint8_t *mem2hex (uint8_t *mem, uint8_t *buf, int count);
static uint8_t *hex2mem (uint8_t *buf, uint8_t *mem, int count);
static int      hex2int (uint8_t **ptr, int *intValue);

static void     real_putDebugChar (int ch, int requestfd);
static int      real_getDebugChar (int requestfd);
#define putDebugChar(ch) real_putDebugChar(ch, requestfd)
#define getDebugChar() real_getDebugChar(requestfd)


static const char hexchars[]="0123456789abcdef";

static int echo = 1; /* should we echo chars sent with putDebugChar? */


#define BUFSZ 1024
#define MEMBUFSZ (BUFSZ/2)

/* Read GDB commands from "requestfd", and do the appropriate
 * operation on the backend */
/* XXX this won't work if an otherwise valid request comes in as multiple 
 * read(2) returns. */
void
stub_loop (int requestfd, backend_ops *ops, void *be)
{
    uint8_t buf[BUFSZ];
    uint8_t membuf[MEMBUFSZ];
    uint8_t *ptr;
    int addr;
    int length;
    int rbytes;
    int sigval;
    int regbytes;
    uint8_t *registers = NULL;
    uint8_t *start;

    /* find out what we died from */
    sigval = ops->sigval (be);
    sigval &= 0xff;

    while (1)
    {
	rbytes = read (requestfd, buf, BUFSZ);
	if (rbytes < 0)
	    die_errno ("read from client");

	if (rbytes == 0)
	{
	    if (ops->finished)
		ops->finished (be);
	    return;
	}

	buf[MIN (rbytes, BUFSZ-1)] = '\000';
	if (verbose > 1)
	    printf ("<- %s\n", buf);

	start = getpacket (buf, rbytes, requestfd);
	if (!start)
	    continue;

	/* work out what we need to do, and format the reply into buf */
	switch (*start) {

	case '?' : /* signal information */
	    buf[0] = 'S';
	    buf[1] =  hexchars[sigval >> 4];
	    buf[2] =  hexchars[sigval % 16];
	    buf[3] = 0;
	    break;

	case 'd' : /* remote debug flag: unsupported */
	    buf[0] = 0;
	    break;

	case 'g' : /* return the value of the CPU registers */
	    registers = ops->regdump (be, &regbytes);
	    mem2hex (registers, buf, regbytes);
	    free(registers);
	    registers = NULL;
	    break;

	case 'G' : /* set the value of the CPU registers - just return OK */
#if 0 /* Not much use unless the backend can support it */
//	    if (!registers)
//		registers = ops->regdump (be, &regbytes);
//	    hex2mem (start + 1, registers, regbytes);
	    strcpy (buf, "OK");
#else
	    strcpy (buf, "E01");
#endif
	    break;

	case 'm' : /* mAA..AA,LLLL  Read LLLL bytes at address AA..AA */
	    /* TRY TO READ %x,%x.  IF SUCCEED, SET PTR = 0 */
	    ptr = start + 1;
	    if (hex2int (&ptr, &addr))
		if (*(ptr++) == ',')
		    if (hex2int (&ptr, &length))
		    {
			if (length > MEMBUFSZ)
			{
			    warn ("asked to read %d bytes, "
				  "but membuf is only %d bytes long",
				  length, MEMBUFSZ);
			}
			else if (ops->memread (be, membuf, addr, length))
			{
			    mem2hex (membuf, buf, length);
			    ptr = 0;
			}
		    }
	    if (ptr)
		strcpy (buf, "E01");
	    break;

	    /* MAA..AA,LLLL: Write LLLL bytes at address AA.AA return OK */
	case 'M' :
	    /* TRY TO READ '%x,%x:'.  IF SUCCEED, SET PTR = 0 */
	    ptr = start + 1;
	    if (hex2int (&ptr, &addr))
		if (*(ptr++) == ',')
		    if (hex2int (&ptr, &length))
			if (*(ptr++) == ':')
			{
			    if (length > MEMBUFSZ)
			    {
				warn ("asked to write %d bytes, "
				      "but membuf is only %d bytes long",
				      length, MEMBUFSZ);
			    }
			    else
			    {
				hex2mem (ptr, membuf, length);
				if (ops->memwrite (be, membuf, addr, length))
				    strcpy (buf, "OK");
				else
				    strcpy (buf, "E03");
				ptr = 0;
			    }
			}
	    if (ptr)
	    {
		strcpy (buf, "E02");
	    }
	    break;

	    /* cAA..AA   Continue at address AA..AA(optional) */
	case 'c' :
	    ptr = start + 1;
	    if (ops->step)
	    {
		if (hex2int (&ptr, &addr))
		    sigval = ops->cont (be, addr, TRUE);
		else
		    sigval = ops->cont (be, addr, FALSE);
		sprintf (buf, "S%02x", sigval);
	    } else {
		sprintf(buf, "E01");
		warn ("'Continue' unimplemented in this backend");
	    }
	    break;

	    /* sAA..AA   Step one instruction from AA..AA(optional) */
	case 's' :
	    ptr = start + 1;
	    if (ops->step)
	    {
		if (hex2int (&ptr, &addr))
		    sigval = ops->step (be, addr, TRUE);
		else
		    sigval = ops->step (be, addr, FALSE);
		sprintf (buf, "S%02x", sigval);
	    } else {
		sprintf(buf, "E01");
		warn ("'Single-step' unimplemented in this backend");
	    }
	    break;

	case 'k' : /* kill the program */
	    /* need to ack now, since we're about to exit */
	    write (requestfd, "+", 1);  /* don't care if this fails */
	    if (ops->kill)
		ops->kill (be);
	    return;
	    break;

	    /* stuff we ignore: */
	case 'H': /* thread control */
	    /* ignore both Hc and Hg commands  */
	    if (start[1] == 'c' || start[1] == 'g') {
		strcpy (buf, "OK");
	    } else {
		strcpy (buf, "E01");
		warn ("unknown thread command '%s'", start);
	    }
	    break;

	case 'q': /* query */
	    /* we don't support offset queries, so return the empty string */
	    if (!strcmp (start, "qOffsets")) {
		buf[0] = 0;
	    } else {
		strcpy (buf, "E02");
		warn ("unhandled query comamnd '%s'", start);
	    }
	    break;

	case 'C':  /* continue / step with signal */
	case 'S':
	    /* I'm sorry, Dave, I can't do that... */
	    buf[0] = 0;
	    break;

	case 'D': /* Detach */
	    if (ops->finished)
		ops->finished (be);
	    strcpy (buf, "OK");
	    putpacket (buf, requestfd);	    
	    return;
	    break;

	default:
	    warn ("unhandled GDB request '%s'", start);
	    buf[0] = 0;  /* we are required to send back the empty string */
	    break;

	} /* end switch */

	/* reply to the request */
	putpacket (buf, requestfd);
    }
}



/* Scan for the sequence $<data>#<checksum>
 * Return pointer to the data, and convert the '#' into '\000'
 * Alternatively, return NULL if there's no valid packet in the buffer.
 */
static uint8_t *
getpacket (uint8_t *buffer, int buflen, int requestfd)
{
  uint8_t checksum;
  uint8_t xmitcsum;
  int ret;
  uint8_t ch;
  uint8_t *start;

  /* wait around for the start character, ignore all other characters */
  while (buflen > 0 && (*buffer & 0x7f) != '$')
  {
      buffer++;
      buflen--;
  }

  /* do we have enough buffer left for a plausable looking packet? */
  if (buflen < 5)
      return NULL;

  buffer++; /* step past the '$' */
  buflen--;

  start = buffer;

  checksum = 0;
  xmitcsum = -1;

  /* now skip forward until a # or end of buffer is found, calculating
   * the checksum as we go */
  do {
      ch = *buffer++ & 0x7f;
      buflen--;
      if (ch == '#') break;
      checksum = checksum + ch;
  } while (buflen > 0);

  /* did we go off the end of the buffer? */
  if (ch != '#')
      return NULL;

  /* blow away the # */
  *(buffer-1) = '\000';

  if (buflen < 2)
      return NULL;
  xmitcsum = hex(*buffer++ & 0x7f) << 4;
  xmitcsum += hex(*buffer++ & 0x7f);
  buflen -= 2;

  if (checksum != xmitcsum)
  {
      putDebugChar ('-');
      return NULL;
  }

  /* successful transfer */
  putDebugChar ('+');

  /* if a sequence char is present, reply the sequence ID */
  if (start[2] == ':')
  {
      ret = write (requestfd, start, 2);
      if (ret != 2)
	  die_errno ("write of client's sequence number failed");
      if (verbose > 1)
	  printf ("-> %.2s\n", start);
      /* remove sequence chars from buffer */
      start += 3;
  }

  return start;
}




/* send the packet in buffer */
static void
putpacket (uint8_t *buffer, int requestfd)
{
  int count;
  int ret;
  uint8_t ch;
  uint8_t checksum;

  /*  $<packet info>#<checksum>. */

  checksum = 0;
  count    = 0;
  while ((ch = buffer[count]))
  {
      checksum += ch;
      count += 1;
  }

  echo = 0;
  do {
      putDebugChar ('$');

      if (verbose > 1)
	  printf ("-> $");

      if (count)
      {
	  ret = write (requestfd, buffer, count);
	  if (ret != count)
	      die_errno ("write data to client");
	  if (verbose > 1)
	      printf ("%s", buffer);
      }

      putDebugChar ('#');
      putDebugChar (hexchars[checksum >> 4]);
      putDebugChar (hexchars[checksum % 16]);

      if (verbose > 1)
	  printf ("#%c%c\n", hexchars[checksum >> 4], hexchars[checksum % 16]);

  } while ((getDebugChar() & 0x7f) != '+');
  echo = 1;
}



static void
real_putDebugChar (int ch, int requestfd)
{
    int ret;
    uint8_t c;

    c = ch;

    ret = write (requestfd, &c, 1);  /* failed checksum */
    if (ret != 1)
	die_errno ("write '%c' to client", ch);
    if (echo && verbose > 1)
	printf ("-> %c\n", ch);
}


static int
real_getDebugChar (int requestfd)
{
    int ret;
    uint8_t c;

    ret = read (requestfd, &c, 1);
    if (ret == 0)
	die ("early EOF on read from client");
    if (ret < 0)
	die_errno ("read from client");

    if (verbose > 1)
	printf ("<- %c\n", (int)c);

    return (int)c;
}


/**************************************************************/
/* Hex conversion routines */

static int
hex (char ch)
{
  if ((ch >= 'a') && (ch <= 'f')) return (ch-'a'+10);
  if ((ch >= '0') && (ch <= '9')) return (ch-'0');
  if ((ch >= 'A') && (ch <= 'F')) return (ch-'A'+10);
  return (-1);
}


/* convert the memory pointed to by mem into hex, placing result in buf */
/* return a pointer to the last char put in buf (null) */
static uint8_t *
mem2hex (uint8_t *mem, uint8_t *buf, int count)
{
      int i;
      uint8_t ch;

      for (i=0; i < count; i++)
      {
          ch = *mem++;
          *buf++ = hexchars[ch >> 4];
          *buf++ = hexchars[ch % 16];
      }

      *buf = 0;
      return buf;
}

/* convert the hex array pointed to by buf into binary to be placed in mem */
/* return a pointer to the character AFTER the last byte written */
uint8_t *
hex2mem (uint8_t *buf, uint8_t *mem, int count)
{
      int i;
      uint8_t ch;

      /* Doesn't cope well with non-hex crap in buffer, but then it
       * should have failed the checksum earlier. */
      for (i=0; i < count; i++)
      {
          ch = hex (*buf++) << 4;
          ch = ch + hex (*buf++);
          *mem++ = ch;
      }

      return mem;
}

static int
hex2int (uint8_t **ptr, int *intValue)
{
    int numChars = 0;
    int hexValue;

    *intValue = 0;

    while (**ptr)
    {
        hexValue = hex(**ptr);
        if (hexValue >=0)
        {
            *intValue = (*intValue <<4) | hexValue;
            numChars ++;
        }
        else
            break;

        (*ptr)++;
    }

    return numChars;
}

/* End of stub.c */
